/*-
 * Copyright (c) 2002 Norikatsu Shigemura
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#define	DEV_SND_MAJOR			30
#define	ENV_DSPDEV			"MOZILLA_DSP_DEVICE"
#define	ENV_DISABLER			"MOZILLA_HACK_DISABLE"
#define	IDENT				"mozilla-flash-glue"

#include <sys/types.h>

#ifdef	WITH_ESOUND
#include <esd.h>
#endif
#include <dlfcn.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>

#if __FreeBSD_version < 400000
#	include <sys/sysctl.h>
#	include <machine/soundcard.h>
#	define	ioctl(d, r, a)		_thread_sys_ioctl(d, r, void *args)
#	define	open(p, f, m)		_thread_sys_open(p, f, void *args)
#	define	_ioctl(d, r, ...)	syscall(SYS_ioctl,  d, r ,##__VA_ARGS__)
#	define	_open(p, f, ...)	syscall(SYS_fhopen, d, r ,##__VA_ARGS__)
#	define	va_list			int
#	define	va_start(a, r)		
#	define	va_arg(a, t)		(args)
#	define	va_end(a)		
#else
#	include	<stdarg.h>
#	include	<sys/soundcard.h>
#endif

#include "linux-fcntl.h"
#include "linux-ioctl.h"

#define	DEBUG
#ifdef	DEBUG
#	define	dprintf(str, ...)	fprintf(stderr, str "\n" ,##__VA_ARGS__)
#else
#	define	dprintf(str, ...)
#endif


static unsigned long
__ioctl_request_mask__ = 1;		/* available: 0 or (unsigned long) -1 */


static void
__ioctl_request_check__(void)
{
	struct utsname u_name;

	if(  getenv(ENV_DISABLER) != NULL  )  {
		__ioctl_request_mask__ = 0;			/* do nothing */

	}  else if(  uname(&u_name) < 0
		|| strncmp(u_name.sysname, "Linux", SYS_NMLN) == 0  )  {
		dprintf("OS=Linux/'%s'", u_name.sysname);	/* XXX: wrong */
		__ioctl_request_mask__ = 0;			/* do nothing */

	}  else  {
		dprintf("OS=FreeBSD/'%s'", u_name.sysname);	/* XXX: wrong */
		__ioctl_request_mask__ = (unsigned long) -1L;	/* do change  */
		unsetenv("LD_PRELOAD");

	}

	dprintf("__ioctl_request_mask__ = %08X", __ioctl_request_mask__);
}


#ifdef	WITH_ESOUND
static int	esdfd    =   -1;
static int	settings =    0;
static int	done     =    0;
#endif

int
open(const char *path, int flags, .../* mode_t mode */)
{
	int		ret;
	va_list		ap;

	va_start(ap, flags);

#ifdef	WITH_ESOUND

	if(  strncmp(path, DSP_DEVICE, sizeof(DSP_DEVICE) - 1) == 0  )  {
		ret = esdfd = esd_open_sound(NULL);
		settings = done = 0;
		dprintf("ESOUND(%s => %d)", path, ret);
	}  else  {
		if(  flags & O_CREAT  )  {
			ret = _open(path, flags, va_arg(ap, mode_t));
		}  else  {
			ret = _open(path, flags);
		}
		dprintf("OPEN(%d) %s / %08X", ret, path, flags);
	}

#else

	if(  strncmp(path, DSP_DEVICE, sizeof(DSP_DEVICE) - 1) == 0  )  {
		const char	*name;

		if(  ( name = getenv(ENV_DSPDEV) )  !=  NULL  )  {
			struct stat st;
			if(  stat(name, &st) == 0 
			 && ( S_ISCHR(st.st_mode) || S_ISBLK(st.st_mode) )
			 && major(st.st_rdev) == DEV_SND_MAJOR
			  )  {
				path = name;
			}
		}
	}

	if(  flags & O_CREAT  )  {
		ret = _open(path, flags, va_arg(ap, mode_t));
	}  else  {
		ret = _open(path, flags);
	}
	dprintf("OPEN(%d) %s / %08X)", ret, path, flags);

#endif

	va_end(ap);

	return ret;
}


int
ioctl(int d, unsigned long request, ...)
{
	int			ret;
	va_list			ap;
#ifdef	WITH_ESOUND
	static int		speed;
	static esd_format_t	fmt = ESD_STREAM | ESD_PLAY | ESD_MONO;
#endif

	va_start(ap, request);

	if(  __ioctl_request_mask__ == 1  )
		__ioctl_request_check__();

	switch(  request & __ioctl_request_mask__  )  {

	case	LINUX_SNDCTL_DSP_RESET:
		dprintf("LINUX_SNDCTL_DSP_RESET");
#ifdef	WITH_ESOUND
		ret = 0;				/* do nothing */
#else
		ret = _ioctl(d, SNDCTL_DSP_RESET);
#endif
		break;

	case LINUX_SNDCTL_DSP_SPEED:
		dprintf("LINUX_SNDCTL_DSP_SPEED");
#ifdef	WITH_ESOUND
		speed = * va_arg(ap, int*);
		settings |= 2;
		ret = 0;
#else
		ret = _ioctl(d, SNDCTL_DSP_SPEED,       va_arg(ap, int*));
#endif
		break;

	case LINUX_SNDCTL_DSP_STEREO:
		dprintf("LINUX_SNDCTL_DSP_STEREO");
#ifdef	WITH_ESOUND
		fmt &= ~ESD_MONO;
		fmt |= *va_arg(ap, int*) ? ESD_STEREO : ESD_MONO;
		ret = 0;
#else
		ret = _ioctl(d, SNDCTL_DSP_STEREO,      va_arg(ap, int*));
#endif
		break;

/*
	case LINUX_SNDCTL_DSP_GETBLKSIZE:
		dprintf("LINUX_SNDCTL_DSP_GETBLKSIZE");
#ifdef	WITH_ESOUND
		*va_arg(ap, int*) = ESD_BUF_SIZE;
		ret = 0;
#else
		ret = _ioctl(d, SNDCTL_DSP_GETBLKSIZE,  va_arg(ap, int*));
#endif
		break;
*/

	case LINUX_SNDCTL_DSP_SETFMT:
		dprintf("LINUX_SNDCTL_DSP_SETFMT");
#ifdef	WITH_ESOUND
		fmt |= (*va_arg(ap, int*) & 0x30) ? ESD_BITS16 : ESD_BITS8;
		settings |= 1;
		ret = 0;
#else
		ret = _ioctl(d, SNDCTL_DSP_SETFMT,      va_arg(ap, int*));
#endif
		break;

	case LINUX_SNDCTL_DSP_SETFRAGMENT:
		dprintf("LINUX_SNDCTL_DSP_SETFRAGMENT");
#ifdef	WITH_ESOUND
		ret = 0;				/* do nothing */
#else
		ret = _ioctl(d, SNDCTL_DSP_SETFRAGMENT, va_arg(ap, int*));
#endif
		break;

	/* XXX: assume that 'struct audio_buf_info' is same */
	case LINUX_SNDCTL_DSP_GETOSPACE:
		dprintf("LINUX_SNDCTL_DSP_GETOSPACE");
#ifdef	WITH_ESOUND
		{
			audio_buf_info *abi = va_arg(ap, linux_audio_buf_info*);
			abi->bytes      = ESD_BUF_SIZE;
			abi->fragments  = 1;
			abi->fragstotal = 1;
			abi->fragsize   = ESD_BUF_SIZE;
		}
		ret = 0;
#else
		ret = _ioctl(d, SNDCTL_DSP_GETOSPACE,   va_arg(ap, linux_audio_buf_info*));
#endif
		break;

	case LINUX_SNDCTL_DSP_NONBLOCK:
		dprintf("LINUX_SNDCTL_DSP_NONBLOCK");
#ifdef	WITH_ESOUND
		ret = 0;				/* do nothing */
#else
		ret = _ioctl(d, SNDCTL_DSP_NONBLOCK);
#endif
		break;

	case LINUX_SNDCTL_DSP_GETCAPS:
		dprintf("LINUX_SNDCTL_DSP_GETCAPS");
#ifdef	WITH_ESOUND
		*va_arg(ap, int*) = 0;
		ret = 0;
#else
		ret = _ioctl(d, SNDCTL_DSP_GETCAPS,     va_arg(ap, int*));
#endif
		break;

#ifdef	DEBUG
	case 0x8004667e:	/* XXX: FreeBSD	sys/filio.h	FIONREAD */
	case 0x4004667f:	/* XXX: FreeBSD	sys/filio.h	FIONBIO  */
	case 0x402C7413:	/* XXX: FreeBSD sys/ttycom.h	TIOCGETA */
		ret = _ioctl(d, request, va_arg(ap, void*));
		break;
#endif

	default:
		dprintf("UNKOWN: %08X", request);
		ret = _ioctl(d, request, va_arg(ap, void*));
		break;

	}

#ifdef	WITH_ESOUND
	if(  settings == 3 && !done  )  {
		int proto = ESD_PROTO_STREAM_PLAY;
		dprintf("ESOUND SETUP");
		done = 1;
		if(  write(esdfd, &proto, sizeof (proto)) != sizeof (proto)  )
			return -1;
		if(  write(esdfd, &fmt,   sizeof (fmt)  ) != sizeof (fmt)    )
			return -1;
		if(  write(esdfd, &speed, sizeof (speed)) != sizeof (speed)  )
			return -1;
		if(  write(esdfd, IDENT,  ESD_NAME_MAX  ) != ESD_NAME_MAX    )
			return -1;

		fmt   = ESD_STREAM | ESD_PLAY | ESD_MONO;
		speed = 0;
		dprintf("ESOUND SETUPED(%d)", ret);
	}
#endif

	va_end(ap);

	return ret;
}

int
close(int d)
{
	int ret;

#ifdef	WITH_ESOUND

	if(  esdfd == d  )  {
		dprintf("ESDCLOSE(%d)", d);
		esdfd = -1;
		settings = done = 0;
		ret = esd_close(d);
	}  else  {
		dprintf("CLOSE(%d)", d);
		ret = _close(d);
	}

#else

	ret = _close(d);

#endif

	return ret;
}


#if __GNUC__ > 2

/* 
 * __builtin_vec_new, __builtin_vec_delete are in gcc 2.95.4's libgcc.a.
 * But there are not in new gcc (> 3.0?).  
 */

void *
__builtin_vec_new(size_t sz /* XXX: std::nothrow_t& nothrow */)
{
	void *p;

	if (sz == 0)
		sz = 1;

	p = malloc(sz);

	/* XXX: ERROR HANDLER is not available */

	return p; 
}

void
__builtin_vec_delete(void *ptr)
{
	if (ptr)
		free (ptr);
}

#endif
